import java.awt.Rectangle;
import java.util.Properties;

import ij.*;
import ij.gui.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import mmorpho.*;

/*
* @version  1.5 19 Nov 2006 
* 				- added support for the brightness channel of RGB images
* 			1.0 7 Oct 2006

*            
* @author Dimiter Prodanov
* 		   Catholic University of Louvain
*
*
* @contents This plugin performs the basic morphologic operations on 32-bit float images
*      erosion, dilation, opening and closing with several types of structuring elements.
*      It is build upon the FloatSE class
*
*      The develpoment of this alogorithm was inspired by the book of Jean Serra
*      "Image Analysis and Mathematical Morphology"
*
* @license This library is free software; you can redistribute it and/or
*      modify it under the terms of the GNU Lesser General Public
*      License as published by the Free Software Foundation; either
*      version 2.1 of the License, or (at your option) any later version.
*
*      This library is distributed in the hope that it will be useful,
*      but WITHOUT ANY WARRANTY; without even the implied warranty of
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*       Lesser General Public License for more details.
*
*      You should have received a copy of the GNU Lesser General Public
*      License along with this library; if not, write to the Free Software
*      Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
public class FloatMorphology_ implements PlugInFilter,  Constants {
	ImagePlus imp;
    public String kernelText = " 0 0 0 0 0\n 0 0 255 0 0\n 0 255 255 255 0\n 0 0 255 0 0\n 0 0 0 0 0\n";
    private static final String R="SE_r", SHOW="show_SE", SETYPE="SE_type", OPER="MOper";
    boolean canceled = true;
    private static final int[] offset=OFFSET0;
    public FloatSE se, minus_se, plus_se, down_se, up_se;
    FloatMorphoProcessor fmp;
    ImageWindow win;
    private static float radius=(float)Prefs. getDouble(R,1);
    private static int options=Prefs.getInt(SETYPE,0);
    private static boolean showoptions=Prefs.getBoolean(SHOW,false);
    private static int morphoptions=Prefs.getInt(OPER,0);
    public final static String[] strelitems={"circle","diamond","square","hor line","ver line","2p h","2p v","free form"};
    public final static int[] constitems={CIRCLE,DIAMOND,SQARE,HLINE,VLINE,HPOINTS,VPOINTS,FREE};
    public final static String[] morphitems={"erode","dilate","open","close"};
    //,"fast erode","fast dilate","fast open","fast close"};
    
    public final static int ERODE=0, DILATE=1,OPEN=2,CLOSE=3,
    FERODE=4,FDILATE=5,FOPEN=6,FCLOSE=7;
    private Roi roi;
    boolean isLineRoi;
    int slice=0;
    private boolean seshown=false;
	/* (non-Javadoc)
	 * @see ij.plugin.filter.PlugInFilter#run(ij.process.ImageProcessor)
	 */
	public void run(ImageProcessor ip) {
		int shift=1;
        // width=(int)(2*radius+2*shift);
        if (IJ.escapePressed())
        {IJ.beep(); return;}
        FloatStatistics fs=null;
        if (ip instanceof FloatProcessor){
        	 fs=new FloatStatistics ((FloatProcessor)ip);
       
        } else if (ip instanceof ColorProcessor){
        	FloatProcessor brightness=
  			((ColorProcessor) ip).getBrightness();
        	 fs=new FloatStatistics (brightness);

        }
        double cmin=fs.histMin;
        double cmax=fs.histMax;
        //IJ.log("histMin: "+ cmin+" histMax: "+ cmax);
        //IJ.log( "options SE "+strelitems[options]+ " "+ constitems[options]);
        int eltype=constitems[options];
        if (eltype==FREE) {
        	 se=inputSE();
        } else {
            shift=1;
            //se=new FloatSE(eltype,  shift,  radius, cmin, cmax);
            se=new FloatSE(eltype, shift, radius, (float)cmin, (float)cmax);
           
            //FloatSE(int type, int shift, float radius, float cmin, float cmax)
        }
        
        
        if (se!=null) {
             fmp=new FloatMorphoProcessor(se);
             if ((showoptions) && (!seshown)){

                showStrEl(se, "SE r=" +radius);
                
             }
	         slice++;   
	         //IJ.showStatus("Doing slice " + slice);
	         if (slice>1)
	             IJ.showStatus(imp.getTitle()+" : "+slice+"/"+imp.getStackSize());
	           // ip.snapshot();
	
	         Rectangle r=ip.getRoi();
	         
	         
	         if (r==null){
	        	doOptions((FloatProcessor)ip, fmp, morphoptions);
	         } // end if
	         else if (!isLineRoi) {
	        	 //ImageProcessor ipmask=null;
	        	 if (ip instanceof FloatProcessor){
	        		 ImageProcessor  ipmask = getMask((FloatProcessor)ip,r);
	        		  doOptions((FloatProcessor)ipmask, fmp, morphoptions);
	        		 ip.insert(ipmask,r.x,r.y);
	                	
	                	ip.reset(ip.getMask());
	        	 }
	        	 else if (ip instanceof ColorProcessor){
	        		 FloatProcessor brightness=
	     			((ColorProcessor) ip).getBrightness();
	        		 ImageProcessor  ipmask = getMask(brightness,r);
	        		  doOptions((FloatProcessor)ipmask, fmp, morphoptions);
	        		  brightness.insert(ipmask,r.x,r.y);

	             	((ColorProcessor) ip).setBrightness(brightness);
	         	}
//	        	ImagePlus rimp=new ImagePlus();
//	        	rimp.setProcessor("roi", ipmask);
//	        	rimp.show();
	        	  	
            	
           
            	
	         } // end if
	        
            if (slice== imp.getImageStackSize())
        	   imp.updateAndDraw();
        }

	}
	
	private void doOptions(FloatProcessor ip,  FloatMorphoProcessor fmp, int morphoptions){
		fmp.calcStat(ip);
		
		switch (morphoptions) {
        
        case ERODE: {
            fmp.erode(ip);
            break;}
        case DILATE:{
            fmp.dilate(ip);
            break;}
        case OPEN:{
            fmp.open(ip);
            break;
        }
        case CLOSE:{
            fmp.close(ip);
            break;
        }
        
    /*    case FERODE: {
        	if ((se.getType()==HLINE) || (se.getType()==VLINE)) {
        		mp.LineErode(ip);}
        	else {
        		mp.fastErode(ip);
        	}
            break;
        }
        case FDILATE:{
        	if ((se.getType()==HLINE) || (se.getType()==VLINE)) {
            	mp.LineDilate(ip);}
        	else {
        		mp.fastDilate(ip);
            	}
            break;
        }
        case FOPEN:{
        	if ((se.getType()==HLINE) || (se.getType()==VLINE)) {
        		mp.LineErode(ip);
        		mp.LineDilate(ip);
            }
        	else {
        		mp.fopen(ip);
        	}
            break;
        }
        case FCLOSE:{
        	if ((se.getType()==HLINE) || (se.getType()==VLINE)) {
            	mp.LineDilate(ip);
            	mp.LineErode(ip);
            	}
        	else {
        		mp.fclose(ip);
        	}
            break;
        }*/
 
    } // switch
    	
    }

	/* (non-Javadoc)
	 * @see ij.plugin.filter.PlugInFilter#setup(java.lang.String, ij.ImagePlus)
	 */
	public int setup(String arg, ImagePlus imp) {
		this.imp=imp;
        
        //IJ.register(GrayMorphology_.class);
        if (arg.equals("about")){
            showAbout();
            return DONE;
        }
        else {
            if (imp!=null) {
                win = imp.getWindow();
                win.running = true;
                isLineRoi=processibleRoi(imp);
            }

            
            if(IJ.versionLessThan("1.35") || !showDialog(imp)) {
                return DONE;
            }
            else {
                return DOES_32+DOES_RGB+DOES_STACKS;
            }
        }
	}
	
	
	public boolean processibleRoi(ImagePlus imp) {
    	roi = imp.getRoi();
        boolean ret=(roi!=null && (roi.getType()==Roi.LINE || 
        						 roi.getType()==Roi.POLYLINE ||
        						 roi.getType()==Roi.ANGLE ||
        						 roi.getType()==Roi.FREELINE ||
        						 roi.getType()==Roi.POINT
        						 )
        		   );
        //IJ.log("roi type "+ isLineRoi);
        return ret;
    }
    
    
    boolean showDialog(ImagePlus imp)   {
        
        if (imp==null) return true;
        GenericDialog gd=new GenericDialog("Parameters");
        
        // Dialog box for user input
        gd.addMessage("This plugin performs morphology operators on float images\n");
        
        gd.addNumericField("Radius of the structure element (pixels):", radius, 1);
        
        gd.addChoice("Type of structure element", strelitems, strelitems[options]);
        gd.addCheckbox("Show mask", showoptions);
        gd.addChoice("Operator", morphitems, morphitems[morphoptions]);

        gd.showDialog();
        radius=(float)gd.getNextNumber();
        options=gd.getNextChoiceIndex();
        
        showoptions=gd.getNextBoolean();
        morphoptions=gd.getNextChoiceIndex();

        
        if (gd.wasCanceled())
            return false;
        if (!MTools.validate(radius,2)){
            IJ.showMessage("Invalid Numbers!\n" +
            "Enter floats 0.5 or 1");
            return false;
        }
               
        return true;
    }
    
	
	
    /* Creates a StructureElement
     * from text input; must be delimited
     *
     * @return StructureElement
     */
    private FloatSE inputSE() {
        GenericDialog gd = new GenericDialog("Input Mask", IJ.getInstance());
        gd.addTextAreas(kernelText, null, 10, 30);
        gd.showDialog();
        if (gd.wasCanceled()) {
            canceled = true;
            return null;
        }
        
        kernelText = gd.getNextText();        
        return new FloatSE(kernelText, Float.MIN_VALUE, Float.MAX_VALUE);
    }
    
    /* Extracts the FloatProcessor within a rectangular roi
     * 
     * @param ip
     * @param r
     * @return FloatProcessor
     */
    public FloatProcessor getMask(FloatProcessor ip, Rectangle r) {
    	
     int width = ip.getWidth();
   	 float[] pixels = (float[])ip.getPixels();

   	 
   	 int xloc=(int)r.getX(); int yloc=(int)r.getY();
     int w=(int)r.getWidth();
     int h=(int)r.getHeight();
     //IJ.log("roi length " +w*h);
     float[] mask=new float[w*h];
 
     for (int cnt=0; cnt<mask.length;cnt++) {
        int index=xloc+cnt%w + (cnt/w)*width +yloc*width;
        mask[cnt]=pixels[index];  
              	
     }

     return new FloatProcessor(w, h, mask, ip.getColorModel());

    }
    
    /** displays the StructureElement */
    
    /**
     * @param strel
     * @param Title
     */
    public void showStrEl(FloatSE strel, String Title) {
        int wh=strel.getWidth();
        int hh=strel.getHeight();
        IJ.log("width: "+wh+" height: "+hh);
        //Log(strel.getMask());
        MTools.Log(se.getFvect());
        ImageProcessor fp = new FloatProcessor(wh,hh);
        fp.setPixels(strel.getFmask());
        fp.resetMinAndMax();
        new ImagePlus(Title, fp).show();
        seshown=true;
    }
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub
		new ImageJ();
	}
	
/*     validates the input value
     * only n/2 floats are accepted
     
    private boolean validate( float var, int k){
        float a=k*var;
        int b=(int) (k* var);
        // IJ.log(IJ.d2s(a-b));
        if ((a-b==0)||(var<0))
            return true;
        else return false;
    }*/
    
    
/*    *//** Logs array
     * @param arr the array
     *//*
    public void Log(float[][] a){
        String aStr="";
        // int w=(int) Math.sqrt(a.length);
        int h=a.length;
        for (int i=0;i<h;i++) {
            for (int j=0; j<a[i].length;j++){
                aStr+=a[i][j]+"  ";
            }
            IJ.log(aStr);
            aStr="";
        }
        
    }*/
    /* Saves the current setings of the plugin for further use
     * 
     *
    * @param prefs
    */
   public static void savePreferences(Properties prefs) {
           prefs.put(R, Double.toString(radius));
           prefs.put(SHOW,Boolean.toString(showoptions));
           prefs.put(SETYPE, Integer.toString(options));
           prefs.put(OPER, Integer.toString(morphoptions));
 
   }
    /*------------------------------------------------------------------*/
    void showAbout() {
        IJ.showMessage("Float Morphology version  1.0",
        "This plugin performs the basic morphologic operations on float images \n  "+
        "erosion, dilation, opening and closing with several types of structuring elements.\n" +
      
        "The develpoment of this alogorithm was inspired by the book of Jean Serra \n" +
        "\"Image Analysis and Mathematical Morphology\""
        );
        
    } /* showAbout */

}
